# SPDX-FileCopyrightText: 2024 UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

import os
import time
import traceback
from datetime import datetime, timedelta, timezone
from pathlib import Path

import urllib3

from dyff.client import Client, HttpResponseError
from dyff.schema.platform import *
from dyff.schema.requests import *

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


class bcolors:
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"


ctx = dict()


def create_dyff_client():
    api_key = os.environ.get("DYFF_API_KEY")
    endpoint = os.environ.get("DYFF_API_ENDPOINT")

    if not api_key or not endpoint:
        raise ValueError(
            "DYFF_API_KEY and DYFF_API_ENDPOINT environment variables must be set"
        )

    return Client(api_key=api_key, endpoint=endpoint, insecure=True, timeout=30)


def wait_for_success(get_entity_fn, *, timeout: timedelta):
    return  # Don't need to wait, can queue
    then = datetime.now(timezone.utc)
    while (datetime.now(timezone.utc) - then) < timeout:
        try:
            status = get_entity_fn().status
            if is_status_success(status):
                return
            elif is_status_failure(status):
                raise AssertionError(f"failure status: {status}")
        except HttpResponseError as ex:
            if ex.status_code != 404:
                raise
        time.sleep(1)
    raise AssertionError("timeout")


# Use existing method
def create_new_model_score(client: Client, num_replicas: int):
    for replication in range(num_replicas):
        try:
            print(
                f"Creating {bcolors.UNDERLINE}Model{bcolors.ENDC} ",
                end="",
            )

            model_request = ModelCreateRequest(
                name=f"mock-model-compare-{replication}",
                account="test",
                artifact=ModelArtifact(
                    kind=ModelArtifactKind.Mock,
                ),
                storage=ModelStorage(
                    medium=ModelStorageMedium.Mock,
                ),
                source=ModelSource(
                    kind=ModelSourceKinds.Mock,
                ),
                resources=ModelResources(storage="0"),
            )
            model = client.models.create(model_request)

            print(
                f"{model.id} | Publishing... ",
                end="",
            )
            wait_for_success(
                lambda: client.models.get(model.id),
                timeout=timedelta(minutes=2),
            )
            client.models.publish(model.id, "preview")
            print(f"{bcolors.OKGREEN}Success{bcolors.ENDC}")

            try:
                print(
                    f"Creating {bcolors.UNDERLINE}Inf Svc{bcolors.ENDC} ",
                    end="",
                )
                service_request = InferenceServiceCreateRequest(
                    account=ctx["account"],
                    name=f"mock-llm-svc-compare-{replication}",
                    model=model.id,
                    runner=InferenceServiceRunner(
                        kind=InferenceServiceRunnerKind.MOCK,
                        resources=ModelResources(
                            storage="1Gi",
                            memory="2Gi",
                        ),
                    ),
                    interface=InferenceInterface(
                        endpoint="generate",
                        outputSchema=DataSchema.make_output_schema(
                            DyffDataSchema(
                                components=["text.Text"],
                            ),
                        ),
                        inputPipeline=[
                            SchemaAdapter(
                                kind="TransformJSON",
                                configuration={"prompt": "$.text"},
                            ),
                        ],
                        outputPipeline=[
                            SchemaAdapter(
                                kind="ExplodeCollections",
                                configuration={"collections": ["text"]},
                            ),
                        ],
                    ),
                )
                inferenceservice = client.inferenceservices.create(service_request)

                print(
                    f"{inferenceservice.id} | Publishing... ",
                    end="",
                )
                wait_for_success(
                    lambda: client.inferenceservices.get(inferenceservice.id),
                    timeout=timedelta(minutes=2),
                )
                client.inferenceservices.publish(inferenceservice.id, "preview")
                print(f"{bcolors.OKGREEN}Success{bcolors.ENDC}")

                try:
                    print(
                        f"Creating {bcolors.UNDERLINE}Safety Case{bcolors.ENDC} ",
                        end="",
                    )
                    safetycase_jupyter_notebook_request = AnalysisCreateRequest(
                        account=ctx["account"],
                        method=ctx["method"],
                        scope=AnalysisScope(
                            evaluation=None,
                            dataset=None,
                            inferenceService=inferenceservice.id,
                            model=model.id,
                        ),
                        arguments=[
                            AnalysisArgument(
                                keyword="trueName", value="Hans Sprungfeld"
                            ),
                            AnalysisArgument(
                                keyword="isOutlier",
                                value=("true" if replication == 0 else "false"),
                            ),
                        ],
                        inputs=[
                            AnalysisInput(
                                keyword="cromulence", entity=ctx["measurement"]
                            ),
                        ],
                    )
                    safetycase = client.safetycases.create(
                        safetycase_jupyter_notebook_request
                    )
                    print(
                        f"{safetycase.id} | Publishing... ",
                        end="",
                    )
                    wait_for_success(
                        lambda: client.safetycases.get(safetycase.id),
                        timeout=timedelta(minutes=2),
                    )
                    client.safetycases.publish(safetycase.id, "preview")
                    print(f"{bcolors.OKGREEN}Success{bcolors.ENDC}")

                except Exception as e:
                    print(
                        f"{bcolors.FAIL}Failed{bcolors.ENDC}\nSafety Case failed to create: {e}"
                    )
                    break
            except Exception as e:
                print(
                    f"{bcolors.FAIL}Failed{bcolors.ENDC}\nService failed to create: {e}"
                )
                break
        except Exception as e:
            print(f"{bcolors.FAIL}Failed{bcolors.ENDC}\nModel failed to create: {e}")
            break


def create_score_method(client: Client):

    try:
        account = ctx["account"]
        module_jupyter_notebook_dir = Path(
            Path(__file__).parent.resolve(), "data", "module_jupyter_notebook"
        )
        print("Creating jupyter notebook module ", end="")
        module_jupyter_notebook = client.modules.create_package(
            module_jupyter_notebook_dir,
            account=account,
            name="module_jupyter_notebook",
        )
        client.modules.upload_package(
            module_jupyter_notebook, module_jupyter_notebook_dir
        )
        print(f"module_jupyter_notebook: {module_jupyter_notebook.id}")

    except Exception as e:
        print(f"Failed to create jupyter module: {e}")
        return

    try:
        method_jupyter_notebook_request = MethodCreateRequest(
            name="method_notebook_scores",
            scope=MethodScope.InferenceService,
            description="""*Distribution Score Notebook Method*""",
            implementation=MethodImplementation(
                kind=MethodImplementationKind.JupyterNotebook,
                jupyterNotebook=MethodImplementationJupyterNotebook(
                    notebookModule=module_jupyter_notebook.id,
                    notebookPath="test-notebook.ipynb",
                ),
            ),
            parameters=[
                MethodParameter(keyword="trueName", description="His real name")
            ],
            inputs=[
                MethodInput(kind=MethodInputKind.Measurement, keyword="cromulence"),
            ],
            output=MethodOutput(
                kind=MethodOutputKind.SafetyCase,
                safetyCase=SafetyCaseSpec(
                    name="safetycase_notebook",
                    description="""*Markdown Description*""",
                ),
            ),
            scores=[
                ScoreSpec(
                    name="float_unit_primary",
                    title="Float (primary; with unit)",
                    summary="A float with a unit that is the 'primary' score",
                    minimum=0,
                    maximum=1,
                    unit="MJ/kg",
                    priority="primary",
                    valence="positive",
                ),
                ScoreSpec(
                    name="int",
                    title="Integer",
                    summary="An Integer score",
                    valence="positive",
                    priority="secondary",
                ),
                ScoreSpec(
                    name="int_big",
                    title="Integer (Big)",
                    summary="A big Integer score",
                    valence="positive",
                    priority="secondary",
                ),
                ScoreSpec(
                    name="int_percent",
                    title="Integer (Percentage)",
                    summary="A percentage represented as an integer",
                    valence="positive",
                    priority="secondary",
                    minimum=0,
                    maximum=100,
                ),
                ScoreSpec(
                    name="no_display",
                    title="Not displayed",
                    summary="A score that should not be displayed",
                    valence="negative",
                    priority="secondary",
                ),
                ScoreSpec(
                    name="exp_rate_low",
                    title="Exponential Distribution (Low Rate)",
                    summary="An exponential distribution with a low rate, leading to a gradual decay.",
                    valence="positive",
                    priority="secondary",
                    minimum=0,
                    maximum=10,
                ),
                ScoreSpec(
                    name="exp_rate_high",
                    title="Exponential Distribution (High Rate)",
                    summary="An exponential distribution with a high rate, causing rapid decay.",
                    valence="positive",
                    priority="secondary",
                    minimum=0,
                    maximum=10,
                ),
                ScoreSpec(
                    name="poisson_rate_low",
                    title="Poisson Distribution (Low Rate)",
                    summary="A Poisson distribution with a low average event rate, generating fewer events.",
                    valence="positive",
                    priority="secondary",
                    minimum=0,
                    maximum=20,
                ),
                ScoreSpec(
                    name="poisson_rate_high",
                    title="Poisson Distribution (High Rate)",
                    summary="A Poisson distribution with a high average event rate, generating more events.",
                    valence="positive",
                    priority="secondary",
                    minimum=0,
                    maximum=20,
                ),
                ScoreSpec(
                    name="normal_standard",
                    title="Normal Distribution (Standard)",
                    summary="A normal distribution with a mean of 0 and a standard deviation of 1.",
                    valence="positive",
                    priority="secondary",
                    minimum=-5,
                    maximum=5,
                ),
                ScoreSpec(
                    name="normal_shifted",
                    title="Normal Distribution (Shifted)",
                    summary="A normal distribution with a mean of 5 and a standard deviation of 2.",
                    valence="positive",
                    priority="secondary",
                    minimum=0,
                    maximum=10,
                ),
                ScoreSpec(
                    name="bimodal_close",
                    title="Bi-Modal Distribution (Close Peaks)",
                    summary="A bi-modal distribution with two peaks close to each other, creating moderate separation.",
                    valence="positive",
                    priority="secondary",
                    minimum=-5,
                    maximum=15,
                ),
                ScoreSpec(
                    name="bimodal_separated",
                    title="Bi-Modal Distribution (Separated Peaks)",
                    summary="A bi-modal distribution with two peaks far apart, creating clear separation between clusters.",
                    valence="positive",
                    priority="secondary",
                    minimum=-5,
                    maximum=20,
                ),
                ScoreSpec(
                    name="close_together",
                    title="Close Values Around 50",
                    summary="Values generated close to 50 within an expanded range of ±10.",
                    valence="positive",
                    priority="secondary",
                    minimum=0,
                    maximum=100,
                ),
                ScoreSpec(
                    name="in_middle",
                    title="Values in the Middle",
                    summary="Values generated in the middle of the range, centered around 50.",
                    valence="positive",
                    priority="secondary",
                    minimum=0,
                    maximum=100,
                ),
                ScoreSpec(
                    name="near_min",
                    title="Values Near Minimum",
                    summary="Values generated near the minimum, within the range of 0 to 10.",
                    valence="positive",
                    priority="secondary",
                    minimum=0,
                    maximum=100,
                ),
                ScoreSpec(
                    name="near_max",
                    title="Values Near Maximum",
                    summary="Values generated near the maximum, within the range of 90 to 100.",
                    valence="positive",
                    priority="secondary",
                    minimum=0,
                    maximum=100,
                ),
                ScoreSpec(
                    name="extreme_outliers",
                    title="Extreme Outliers",
                    summary="Extreme outlier values that are far outside the typical range.",
                    valence="positive",
                    priority="secondary",
                    minimum=0,
                    maximum=100,
                ),
                ScoreSpec(
                    name="same_values",
                    title="All Values Exactly the Same",
                    summary="All values are exactly the same, uniform data for testing.",
                    valence="positive",
                    priority="secondary",
                    minimum=0,
                    maximum=100,
                ),
            ],
            modules=[module_jupyter_notebook.id],
            account=account,
        )
        method_jupyter_notebook = client.methods.create(method_jupyter_notebook_request)
        client.methods.publish(method_jupyter_notebook.id, "preview")

        print(f"Created method {method_jupyter_notebook.id}")
    except Exception as e:
        tb = traceback.format_exc()
        print(f"Full stack trace: {tb}")
        print(f"${bcolors.FAIL}Failed to create method {e}{bcolors.ENDC}")


# Main function to execute the script
def main():
    try:
        client = create_dyff_client()
        choice = input(
            "1. Create a new method\n2. Create model/svc and safety case\n> "
        )
        if choice == "1":
            create_score_method(client)
        elif choice == "2":
            num_replicas = int(input("# Instances\n> "))
            ctx["account"] = input("Account ID\n> ")
            ctx["dataset"] = input("Dataset ID\n> ")
            ctx["measurement"] = input("Measurement ID\n> ")
            ctx["method"] = input("Method ID\n> ")
            create_new_model_score(client, num_replicas)
        else:
            print("Bad option")

    except ValueError as e:
        print(f"Error: {e}")
    except Exception as e:
        print(f"An unexpected error occurred: {e}")


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2024 UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

# mypy: disable-error-code="import-untyped"
import sys
from typing import Iterable

import pyarrow
import pydantic

from dyff.schema.base import int32
from dyff.schema.dataset import ReplicatedItem
from dyff.schema.dataset.arrow import arrow_schema


class BlurstCountScoredItem(ReplicatedItem):
    blurstCount: int32() = pydantic.Field(  # type: ignore
        description="Number of times the word 'blurst' is used in the response."
    )
    cromulent: int32() = pydantic.Field(  # type: ignore
        description="Whether the text is cromulent."
    )
    embiggen: str = pydantic.Field(description="Which man to embiggen.")


def blurst_count(
    args: dict[str, str],
    *,
    dataset: pyarrow.dataset.Dataset,
    outputs: pyarrow.dataset.Dataset,
) -> Iterable[pyarrow.RecordBatch]:
    print("stdout message")
    print("stderr message", file=sys.stderr)

    embiggen_arg = args["embiggen"]
    schema = arrow_schema(BlurstCountScoredItem)

    cromulence = {}
    for b in dataset.to_batches(columns=["_index_", "text"]):
        for item in b.to_pylist():
            cromulent = 1 if item["text"].startswith("It was the best of times,") else 0
            cromulence[item["_index_"]] = cromulent

    for b in outputs.to_batches(columns=["_index_", "_replication_", "responses"]):
        batch = []
        for item in b.to_pylist():
            index = item["_index_"]
            replication = item["_replication_"]
            text = item["responses"][0]["text"]
            words = [w.lower() for w in text.split()]
            count = words.count("blurst")
            batch.append(
                BlurstCountScoredItem(
                    _index_=index,
                    _replication_=replication,
                    blurstCount=count,
                    cromulent=cromulence[index],
                    embiggen=embiggen_arg,
                ).dict()
            )
        yield pyarrow.RecordBatch.from_pylist(batch, schema=schema)  # type: ignore[attr-defined]


def raise_error(
    args: dict[str, str],
    *,
    dataset: pyarrow.dataset.Dataset,
    outputs: pyarrow.dataset.Dataset,
) -> Iterable[pyarrow.RecordBatch]:
    print("stdout message")
    print("stderr message", file=sys.stderr)
    raise RuntimeError("deliberate error")



# SPDX-FileCopyrightText: 2024 UL Research Institutes
# SPDX-License-Identifier: Apache-2.0
"""Rubrics related to text processing tasks."""

# mypy: disable-error-code="import-untyped"
from __future__ import annotations

import enum
import logging
from typing import Any, Dict, Iterable, List

import pyarrow
import pydantic

from dyff.schema.base import int32
from dyff.schema.dataset import ReplicatedItem
from dyff.schema.dataset.arrow import arrow_schema, schema_function

from .base import Rubric


class MatchCriterion(str, enum.Enum):
    """The match criteria defined by the SemEval2013 evaluation standard.

    See :class:`MatchResult` for the possible results of matching.

    .. seealso::

      https://www.davidsbatista.net/blog/2018/05/09/Named_Entity_Evaluation/
    """

    #: The prediction is ``correct`` if ``[start, end)`` match exactly and the
    #: tag is correct. Otherwise, the prediction is ``incorrect``.
    strict = "strict"

    #: Tag is ignored. The prediction is ``correct`` if ``[start, end)`` match
    #: exactly. Otherwise, the prediction is ``incorrect``.
    exact = "exact"

    #: Tag is ignored. Partial overlap is scored as ``partial``, exact overlap
    #: is scored as ``correct``, no overlap is scored as ``incorrect``.
    partial = "partial"

    #: Partial overlap is scored as either ``correct`` or ``incorrect`` based on
    #: the predicted tag. No overlap is scored as ``incorrect``.
    type = "type"


class MatchResult(str, enum.Enum):
    """The possible results of a text span matching comparison as defined by the MUC-5
    evaluation standard.

    Note that the semantics of these results depend on which
    :class:`MatchCriterion` is being applied.

    .. seealso::

      https://www.davidsbatista.net/blog/2018/05/09/Named_Entity_Evaluation/
    """

    #: The prediction overlapped with a true entity and was correct.
    correct = "correct"

    #: The prediction overlapped with a true entity but was incorrect.
    incorrect = "incorrect"

    #: The prediction overlapped a true entity partially but not exactly.
    #: Applies to ``MatchCriterion.partial`` only.
    partial = "partial"

    #: A ground-truth entity did not overlap with any predicted entity.
    missing = "missing"

    #: A predicted entity did not overlap with any ground-truth entity.
    spurious = "spurious"


def match_spans(
    predictions: List[Dict[str, Any]], truths: List[Dict[str, Any]]
) -> List[Dict[str, Any]]:
    """Compute matching results for ground-truth and predicted text spans using the
    extended set of matching criteria defined by the SemEval2013 and MUC-5 evaluation
    standards.

    :param predictions: List of predicted spans
    :param truths: List of ground-truth spans
    :return: A two-level dictionary containing match counts, like:
        ``{MatchCriterion: {MatchResult: Count}}``

    .. seealso::

      * https://www.davidsbatista.net/blog/2018/05/09/Named_Entity_Evaluation/
      * https://github.com/davidsbatista/NER-Evaluation
      * https://aclanthology.org/S13-2056.pdf
      * https://aclanthology.org/M93-1007.pdf
    """
    truths = list(sorted(truths, key=lambda span: span["start"]))
    predictions = list(sorted(predictions, key=lambda span: span["start"]))
    ti = 0
    intersected_truths = set()
    results = []

    for p, prediction in enumerate(predictions):
        logging.debug(prediction)
        spurious = True
        for t in range(ti, len(truths)):
            truth = truths[t]
            logging.debug(f"\t{truth}")
            if prediction["end"] < truth["start"]:
                logging.debug("\t\tprediction < truth")
                # ....TTTT
                # PPPP....
                break  # All remaining 'truths' come after 'prediction'
            elif truth["end"] < prediction["start"]:
                logging.debug("\t\ttruth < prediction")
                # TTTT....
                # ....PPPP
                ti = t + 1  # This 'truth' and all previous will never overlap again
            elif (
                prediction["start"] < truth["end"]
                and truth["start"] < prediction["end"]
            ):
                # Intersection
                spurious = False
                intersected_truths.add(t)
                correct_tag = (
                    MatchResult.correct
                    if truth["tag"] == prediction["tag"]
                    else MatchResult.incorrect
                )

                if (
                    truth["start"] == prediction["start"]
                    and truth["end"] == prediction["end"]
                ):
                    logging.debug("\t\texact")
                    # Exact overlap
                    # ..TTTT..
                    # ..PPPP..
                    for criterion in MatchCriterion:
                        row = {
                            "MatchCriterion": criterion,
                            "prediction": p,
                            "truth": t,
                            **{str(r): 0 for r in MatchResult},
                        }
                        if criterion in [MatchCriterion.strict, MatchCriterion.type]:
                            row[str(correct_tag)] = 1
                        else:
                            row[str(MatchResult.correct)] = 1
                        results.append(row)
                else:
                    logging.debug("\t\tpartial")
                    # Partial overlap
                    for criterion in MatchCriterion:
                        row = {
                            "MatchCriterion": criterion,
                            "prediction": p,
                            "truth": t,
                            **{str(r): 0 for r in MatchResult},
                        }
                        if criterion in [MatchCriterion.strict, MatchCriterion.exact]:
                            row[str(MatchResult.incorrect)] = 1
                        elif criterion == MatchResult.partial:
                            row[str(MatchResult.partial)] = 1
                        else:
                            row[str(correct_tag)] = 1
                        results.append(row)
        if spurious:
            logging.debug("\tspurious")
            for criterion in MatchCriterion:
                row = {
                    "MatchCriterion": criterion,
                    "prediction": p,
                    "truth": None,
                    **{str(r): 0 for r in MatchResult},
                }
                row[str(MatchResult.spurious)] = 1
                results.append(row)

    missing = set(range(len(truths))) - intersected_truths
    for criterion in MatchCriterion:
        for t, truth in enumerate(truths):
            if t in missing:
                row = {
                    "MatchCriterion": criterion,
                    "prediction": None,
                    "truth": t,
                    **{str(r): 0 for r in MatchResult},
                }
                row[str(MatchResult.missing)] = 1
                results.append(row)

    return results


class TextSpanMatchesScoredItem(ReplicatedItem):
    MatchCriterion: str = pydantic.Field(
        description=":class:`MatchCriterion` applicable to this record."
    )
    prediction: int32() = pydantic.Field(  # type: ignore
        description="Index of the relevant predicted span within the current instance.",
    )
    truth: int32() = pydantic.Field(  # type: ignore
        description="Index of the relevant ground truth span within the current instance.",
    )
    MatchResultCorrect: bool = pydantic.Field(
        alias="MatchResult.correct",
        description="Indicator of whether the match is ``correct``.",
    )
    MatchResultIncorrect: bool = pydantic.Field(
        alias="MatchResult.incorrect",
        description="Indicator of whether the match is ``incorrect``.",
    )
    MatchResultPartial: bool = pydantic.Field(
        alias="MatchResult.partial",
        description="Indicator of whether the match is ``partial``.",
    )
    MatchResultMissing: bool = pydantic.Field(
        alias="MatchResult.missing",
        description="Indicator of whether the match is ``missing``.",
    )
    MatchResultSpurious: bool = pydantic.Field(
        alias="MatchResult.spurious",
        description="Indicator of whether the match is ``spurious``.",
    )


class TextSpanMatches(Rubric):
    """Computes matches between predicted and ground-truth text spans according to each
    criterion in ``MatchCriterion``."""

    def _build_truth_dict(self, truth_dataset):
        truth = {}
        for b in truth_dataset.to_batches(columns=["_index_", "spans"]):
            for item in b.to_pylist():
                truth[item["_index_"]] = item["spans"]
        return truth

    @property
    def name(self) -> str:
        return "text.TextSpanMatches"

    @property
    @schema_function(arrow_schema(TextSpanMatchesScoredItem))
    def schema(self) -> pyarrow.Schema:
        """The PyArrow schema of the output of applying the Rubric.

        There may be 0 or more rows for each input instance (same ``_index_``).
        Each row indicates which predicted span within that instance overlapped
        with which ground truth span, and how that overlap was scored. For example,
        if the 1st predicted span overlapped with the 2nd ground truth span, then
        there will be a row with ``prediction = 1`` and ``truth = 2``. Spans are
        0-indexed in increasing order of their ``.start`` field. A ``spurious``
        match will have a value for ``prediction`` but not for ``truth``, and a
        ``missing`` match will have a value for ``truth`` but not for
        ``prediction``. Instances for which there are no predicted spans and no
        ground truth spans will not appear in the results.
        """

    def apply(
        self, task_data: pyarrow.datasets.Dataset, predictions: pyarrow.datasets.Dataset
    ) -> Iterable[pyarrow.RecordBatch]:
        truth_dict = self._build_truth_dict(task_data)
        processed = []
        for b in predictions.to_batches(
            columns=["_index_", "_replication_", "responses"]
        ):
            batch = []
            for item in b.to_pylist():
                index = item["_index_"]
                replication = item["_replication_"]
                predicted_spans = item["responses"][0]["spans"]
                truth_spans = truth_dict[index]
                results = match_spans(predicted_spans, truth_spans)
                for result in results:
                    batch.append(
                        {"_index_": index, "_replication_": replication, **result}
                    )
                processed.append(index)
            yield pyarrow.RecordBatch.from_pylist(batch, schema=self.schema)


__all__ = [
    "MatchCriterion",
    "MatchResult",
    "TextSpanMatches",
    "TextSpanMatchesScoredItem",
    "match_spans",
]


def _test():
    import json  # pylint: disable=import-outside-toplevel

    logging.set_verbosity(logging.DEBUG)

    # "Alice Washington and Bob Washington live in Washington National Forest with Wilma Washington"
    #  012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012
    #  0         1         2         3         4         5         6         7         8         9
    truth = [
        {"start": 0, "end": 16, "tag": "PER"},
        {"start": 21, "end": 35, "tag": "PER"},
        {"start": 44, "end": 70, "tag": "LOC"},
        {"start": 76, "end": 92, "tag": "PER"},
    ]

    predictions = [
        {"start": 0, "end": 5, "tag": "PER"},
        {"start": 6, "end": 16, "tag": "ORG"},
        {"start": 36, "end": 40, "tag": "PER"},
        {"start": 21, "end": 35, "tag": "PER"},
        {"start": 21, "end": 35, "tag": "LOC"},
        {"start": 76, "end": 81, "tag": "PER"},
    ]

    print(json.dumps(match_spans(predictions, truth), indent=2))


if __name__ == "__main__":
    _test()



# SPDX-FileCopyrightText: 2024 UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

# mypy: disable-error-code="import-untyped"
from __future__ import annotations

import base64
import json
import shutil
from datetime import datetime, timezone
from pathlib import Path
from typing import Any, Generic, Optional, Type, TypeVar, Union

import ruamel.yaml

from dyff.client import Client, HttpResponseError
from dyff.client.client import (
    DatasetsOperations,
    EvaluationsOperations,
    InferenceservicesOperations,
    InferencesessionsOperations,
    MeasurementsOperations,
    MethodsOperations,
    ModelsOperations,
    ModulesOperations,
    ReportsOperations,
    SafetycasesOperations,
    SchemaObject,
    _ArtifactsMixin,
    _parse_schema_object,
)
from dyff.schema import ids
from dyff.schema.adapters import Adapter, create_pipeline
from dyff.schema.dataset import arrow, binary
from dyff.schema.platform import (
    Analysis,
    AnalysisData,
    Artifact,
    DataSchema,
    Dataset,
    Digest,
    Documentation,
    DyffEntity,
    DyffSchemaBaseModel,
    Entities,
    EntityStatus,
    Evaluation,
    ForeignInferenceService,
    ForeignMethod,
    ForeignModel,
    InferenceInterface,
    InferenceService,
    InferenceSession,
    InferenceSessionAndToken,
    InferenceSessionSpec,
    Measurement,
    Method,
    Model,
    Module,
    Report,
    Resources,
    SafetyCase,
    Score,
)
from dyff.schema.requests import (
    AnalysisCreateRequest,
    DatasetCreateRequest,
    EvaluationCreateRequest,
    InferenceSessionCreateRequest,
    MethodCreateRequest,
    ModelCreateRequest,
    ModuleCreateRequest,
    ReportCreateRequest,
)

from .. import dynamic_import
from .._internal import fqn, upcast
from ..analysis import run_analysis, run_report
from ..workflows import local_evaluation
from . import mocks

_EntityT = TypeVar(
    "_EntityT",
    Dataset,
    Evaluation,
    InferenceService,
    InferenceSession,
    Measurement,
    Method,
    Model,
    Module,
    Report,
    SafetyCase,
)
_EntityOpsT = TypeVar(
    "_EntityOpsT",
    DatasetsOperations,
    EvaluationsOperations,
    InferenceservicesOperations,
    InferencesessionsOperations,
    MeasurementsOperations,
    MethodsOperations,
    ModelsOperations,
    ModulesOperations,
    ReportsOperations,
    SafetycasesOperations,
)


def _namespaced_id(e: DyffEntity) -> str:
    return f"{Resources.for_kind(Entities(e.kind)).value}/{e.id}"


class _Common(Generic[_EntityT, _EntityOpsT]):
    def __init__(
        self,
        platform: DyffLocalPlatform,
        *,
        entity_kind: Entities,
        entity_type: Type[_EntityT],
        remote_ops: Optional[_EntityOpsT],
    ):
        self._platform = platform
        self._entity_kind = entity_kind
        self._entity_type: Type[_EntityT] = entity_type
        self._remote_ops: Optional[_EntityOpsT] = remote_ops

    def get(self, id: str) -> Optional[_EntityT]:
        """Get the database record associated with an entity."""
        filename = f".{self._entity_kind.value.lower()}.json"
        try:
            file = self._platform.entity_path(id) / filename
            return self._entity_type.parse_file(file)
        except FileNotFoundError:
            return None

    def delete(self, id: str) -> None:
        """Set the entity's status to ``"Deleted"``."""
        entity = self.get(id)
        if entity is None:
            raise HttpResponseError(id, status_code=404)
        entity.status = EntityStatus.deleted
        self._platform._commit(entity)

    def fetch(self, id: str) -> _EntityT:
        """Fetch an entity and all associated artifacts from the remote Dyff instance
        associated with the remote client that was provided to the ``DyffLocalPlatform``
        constructor."""
        if not self._remote_ops:
            raise AssertionError(".fetch() requires a remote client")
        entity = self._remote_ops.get(id)
        if entity is None:
            raise KeyError(f"no remote entity with ID {id}")
        if isinstance(self._remote_ops, _ArtifactsMixin):
            self._remote_ops.download(id, self._platform.entity_path(id))
        self._platform._commit(entity)
        return entity  # type: ignore

    def purge(self, id: str) -> None:
        """Remove an entity and all associated artifacts from local storage."""
        entity = self.get(id)
        if entity is None:
            return
        shutil.rmtree(self._platform.entity_path(id))


class _Datasets(_Common[Dataset, DatasetsOperations]):
    def __init__(self, platform: DyffLocalPlatform):
        super().__init__(
            platform,
            entity_kind=Entities.Dataset,
            entity_type=Dataset,
            remote_ops=platform.remote.datasets if platform.remote else None,
        )

    def create(self, request: SchemaObject[DatasetCreateRequest]) -> Dataset:
        """Create a new entity.

        :param request: The entity create request specification.
        :return: A full entity spec with its .id and other system properties set
        """
        id = ids.generate_entity_id()
        request = _parse_schema_object(DatasetCreateRequest, request)
        dataset_dict = request.dict()
        dataset_dict["id"] = id
        dataset_dict["account"] = self._platform.account
        dataset_dict["status"] = EntityStatus.ready
        dataset_dict["creationTime"] = datetime.now(timezone.utc)
        dataset = Dataset.parse_obj(dataset_dict)

        cache_dir = self._platform.entity_path(id)
        cache_dir.mkdir()
        with open(cache_dir / ".dataset.json", "w") as fout:
            fout.write(dataset.json())
        return dataset

    def create_arrow_dataset(
        self, dataset_directory: str, *, account: str, name: str
    ) -> Dataset:
        """Create a Dataset resource describing an existing Arrow dataset.

        Internally, constructs a ``DatasetCreateRequest`` using information
        obtained from the Arrow dataset, then calls ``create()`` with the
        constructed request.

        Typical usage::

            dataset = client.datasets.create_arrow_dataset(dataset_directory, ...)
            client.datasets.upload_arrow_dataset(dataset, dataset_directory)

        :param dataset_directory: The root directory of the Arrow dataset.
        :type dataset_directory: str
        :keyword account: The account that will own the Dataset resource.
        :type account: str
        :keyword name: The name of the Dataset resource.
        :type name: str
        :returns: The complete Dataset resource.
        :rtype: Dataset
        """
        dataset_path = Path(dataset_directory)
        ds = arrow.open_dataset(str(dataset_path))
        file_paths = list(ds.files)  # type: ignore[attr-defined]
        artifact_paths = [
            str(Path(file_path).relative_to(dataset_path)) for file_path in file_paths
        ]
        artifacts = [
            Artifact(
                kind="parquet",
                path=artifact_path,
                digest=Digest(
                    md5=binary.encode(binary.file_digest("md5", file_path)),
                ),
            )
            for file_path, artifact_path in zip(file_paths, artifact_paths)
        ]
        schema = DataSchema(
            arrowSchema=arrow.encode_schema(ds.schema),
        )
        request = DatasetCreateRequest(
            account=account,
            name=name,
            artifacts=artifacts,
            schema=schema,
        )
        return self.create(request)

    def upload_arrow_dataset(self, dataset: Dataset, dataset_directory: str) -> None:
        """Uploads the data files in an existing Arrow dataset for which a Dataset
        resource has already been created.

        Typical usage::

            dataset = client.datasets.create_arrow_dataset(dataset_directory, ...)
            client.datasets.upload_arrow_dataset(dataset, dataset_directory)

        :param dataset: The Dataset resource for the Arrow dataset.
        :type dataset: Dataset
        :param dataset_directory: The root directory of the Arrow dataset.
        :type dataset_directory: str
        """
        if any(artifact.digest.md5 is None for artifact in dataset.artifacts):
            raise ValueError("artifact.digest.md5 must be set for all artifacts")
        for artifact in dataset.artifacts:
            assert artifact.digest.md5 is not None
            file_path = Path(dataset_directory) / artifact.path
            cache_path = self._platform.storage_root / dataset.id / artifact.path
            cache_path.parent.mkdir(parents=True, exist_ok=True)
            with open(file_path, "rb") as fin:
                with open(cache_path, "wb") as fout:
                    fout.write(fin.read())


class _Evaluations(_Common[Evaluation, EvaluationsOperations]):
    def __init__(self, platform: DyffLocalPlatform):
        super().__init__(
            platform,
            entity_kind=Entities.Evaluation,
            entity_type=Evaluation,
            remote_ops=platform.remote.evaluations if platform.remote else None,
        )

    def create(self, request: SchemaObject[EvaluationCreateRequest]) -> Evaluation:
        """Create a new entity.

        :param request: The entity create request specification.
        :return: A full entity spec with its .id and other system properties set
        """
        id = ids.generate_entity_id()
        request = _parse_schema_object(EvaluationCreateRequest, request)
        dataset = self._platform.datasets.get(request.dataset)
        if dataset is None:
            raise HttpResponseError(f"datasets/{request.dataset}", status_code=404)

        evaluation_dict: dict[str, Any] = {}
        evaluation_dict["id"] = id
        evaluation_dict["account"] = request.account
        evaluation_dict["dataset"] = request.dataset
        evaluation_dict["replications"] = request.replications

        if request.inferenceSessionReference is not None:
            session = self._platform.inferencesessions.get(
                request.inferenceSessionReference
            )
            if session is None:
                raise HttpResponseError(
                    f"inferencesessions/{request.inferenceSessionReference}",
                    status_code=404,
                )
            token = self._platform.inferencesessions.token(session.id)
            evaluation_dict["inferenceSessionReference"] = (
                request.inferenceSessionReference
            )
        elif request.inferenceSession is not None:
            service_id = request.inferenceSession.inferenceService
            session_request = InferenceSessionCreateRequest(
                account=request.account, inferenceService=service_id
            )
            session_and_token = self._platform.inferencesessions.create(session_request)
            session = session_and_token.inferencesession
            token = session_and_token.token
        else:
            raise ValueError(
                "must specify one of {inferenceSession, inferenceSessionReference}"
            )

        evaluation_dict["inferenceSession"] = upcast(
            InferenceSessionSpec, session.dict()
        ).dict()
        evaluation = Evaluation.parse_obj(evaluation_dict)
        # TODO: Actually check success / failure
        evaluation.status = EntityStatus.complete

        cache_dir = self._platform.entity_path(evaluation.id)
        cache_dir.mkdir()
        with open(cache_dir / ".evaluation.json", "w") as fout:
            fout.write(evaluation.json())

        inference_client = self._platform.inferencesessions.client(
            session.id, token, interface=session.inferenceService.interface
        )
        session.inferenceService.interface.outputSchema.arrowSchema

        input_dataset = arrow.open_dataset(str(self._platform.entity_path(dataset.id)))

        output_generator = inference_client.run_evaluation(
            input_dataset,
            replications=request.replications,
            id=evaluation.id,
        )
        arrow.write_dataset(
            output_generator,
            output_path=str(self._platform.entity_path(evaluation.id)),
            feature_schema=arrow.decode_schema(
                session.inferenceService.interface.outputSchema.arrowSchema
            ),
        )

        return evaluation

    def import_data(
        self,
        dataset_directory: Path | str,
        *,
        id: Optional[str] = None,
        evaluation_request: Optional[EvaluationCreateRequest] = None,
    ) -> Evaluation:
        """Imports existing data into the local cache under a newly-generated ID. Useful
        when you already have evaluation data and you want to make it available to other
        workflows using the local platform.

        :param dataset_directory: The root directory of the Arrow dataset.
        :type dataset_directory: str
        :returns: The ID of the imported data in the local platform.
        :rtype: str
        """
        dataset_path = Path(dataset_directory)
        ds = arrow.open_dataset(str(dataset_path))
        file_paths = list(ds.files)  # type: ignore[attr-defined]

        replications = set()
        for row in ds.to_table().to_pylist():
            replications.add(row["_replication_"])

        evaluation_dict: dict[str, Any] = {}
        evaluation_dict["id"] = id or ids.generate_entity_id()
        evaluation_dict["account"] = (
            evaluation_request and evaluation_request.account
        ) or ids.null_id()
        evaluation_dict["dataset"] = (
            evaluation_request and evaluation_request.dataset
        ) or ids.null_id()
        evaluation_dict["replications"] = len(replications)

        if evaluation_request and evaluation_request.inferenceSession:
            service_id = evaluation_request.inferenceSession.inferenceService
        else:
            service_id = ids.null_id()
        session_spec = InferenceSessionSpec(
            inferenceService=ForeignInferenceService(
                id=service_id,
                name="",
                account=ids.null_id(),
                interface=InferenceInterface(
                    endpoint="",
                    outputSchema=DataSchema(arrowSchema=arrow.encode_schema(ds.schema)),
                ),
            )
        )
        evaluation_dict["inferenceSession"] = session_spec.dict()
        evaluation = Evaluation.parse_obj(evaluation_dict)
        # TODO: actually check success / failure
        evaluation.status = EntityStatus.complete

        cache_dir = self._platform.entity_path(evaluation.id)
        cache_dir.mkdir()
        with open(cache_dir / ".evaluation.json", "w") as fout:
            fout.write(evaluation.json())

        artifact_paths = [
            str(Path(file_path).relative_to(dataset_path)) for file_path in file_paths
        ]
        artifacts = [
            Artifact(
                kind="parquet",
                path=artifact_path,
                digest=Digest(
                    md5=binary.encode(binary.file_digest("md5", file_path)),
                ),
            )
            for file_path, artifact_path in zip(file_paths, artifact_paths)
        ]
        for artifact in artifacts:
            assert artifact.digest.md5 is not None
            file_path = Path(dataset_directory) / artifact.path
            cache_path = self._platform.entity_path(evaluation.id) / artifact.path
            cache_path.parent.mkdir(parents=True, exist_ok=True)
            with open(file_path, "rb") as fin:
                with open(cache_path, "wb") as fout:
                    fout.write(fin.read())

        return evaluation

    def local_evaluation(self, *, dataset: str, inferencesession: str) -> str:
        """Emulate an Evaluation workflow by feeding data from a local Dataset to an
        InferenceSession running on the Dyff platform.

        The output dataset will have the same schema as the outputs from an
        Evaluation run on the platform, including fields added by the platform
        -- ``_index_``, ``_replication_``, etc.

        The input dataset must be compatible with the canonical Dyff Platform
        dataset schema for the appropriate inference task.

        :param dataset: The ID of a ``Dataset`` in *local storage*.
        :type dataset: str
        :param inferencesession: The ID of an ``InferenceSession`` that is
            *already running* on the *remote Dyff instance*.
        :type inferencesession: ID
        :returns: An ID for the evaluation. This will not correspond to
            an entity in the the local or remote datastores, but it can be used
            to derive the IDs of replications in the output dataset.
        :rtype: str
        """
        client = self._platform._require_client()
        session = client.inferencesessions.get(inferencesession)
        id = ids.generate_entity_id()
        return local_evaluation(
            client,
            session,
            input_dataset_path=self._platform.entity_path(dataset),
            # workflows.local_evaluation() creates an /id subdirectory for the output
            output_dataset_path=self._platform.storage_root,
            id=id,
        )


class _InferenceServices(_Common[InferenceService, InferenceservicesOperations]):
    def __init__(self, platform: DyffLocalPlatform):
        super().__init__(
            platform,
            entity_kind=Entities.InferenceService,
            entity_type=InferenceService,
            remote_ops=platform.remote.inferenceservices if platform.remote else None,
        )

    def create_mock(
        self,
        mock_type: Type[mocks.InferenceServiceMock],
        *,
        account: str,
        id: Optional[str] = None,
        model: Optional[str] = None,
    ) -> InferenceService:
        """Create an InferenceService that uses a mock-up backend.

        The ``mock_type`` can be anything that implements the
        :py:class:`~dyff.audit.local.mocks.InferenceServiceMock` interface
        and is importable locally; you can implement and use your own
        mock services.

        :param mock_type: The type of the mock service implementation.
        :keyword account: The account that should own the InferenceService.
        :keyword id: Will be generated if not given.
        :keyword model: The ID of a model to associate with the service. Since
            this is a mock service, usually the model will also be a mock-up
            (i.e., the ``model.source.kind`` field will be
            :py:attr:`~dyff.schema.platform.ModelSourceKinds.Mock`).
        :return: A full entity spec with its .id and other system properties set
        """
        id = id or ids.generate_entity_id()
        model_spec = self._platform.models.get(model) if model else None

        mock = mock_type()
        service = InferenceService(
            account=account,
            id=id,
            name=".".join(fqn(mock_type)),
            model=upcast(ForeignModel, model_spec.dict()) if model_spec else None,
            interface=mock.interface,
            status=EntityStatus.ready,
        )
        self._platform._commit(service)

        return service


class _InferenceSessions(_Common[InferenceSession, InferencesessionsOperations]):
    def __init__(self, platform: DyffLocalPlatform):
        super().__init__(
            platform,
            entity_kind=Entities.InferenceSession,
            entity_type=InferenceSession,
            remote_ops=platform.remote.inferencesessions if platform.remote else None,
        )

    def create(
        self, request: SchemaObject[InferenceSessionCreateRequest]
    ) -> InferenceSessionAndToken:
        """Create a new entity.

        :param request: The entity create request specification.
        :return: A full entity spec with its .id and other system properties set
        """
        request = _parse_schema_object(InferenceSessionCreateRequest, request)
        id = ids.generate_entity_id()

        service = self._platform.inferenceservices.get(request.inferenceService)
        if service is None:
            raise HttpResponseError(
                f"inferenceservices/{request.inferenceService}", status_code=404
            )

        session = InferenceSession(
            account=request.account,
            id=id,
            inferenceService=upcast(ForeignInferenceService, service.dict()),
            status=EntityStatus.admitted,
        )

        cache_dir = self._platform.entity_path(id)
        cache_dir.mkdir()
        with open(cache_dir / ".inferencesession.json", "w") as fout:
            fout.write(session.json())

        return InferenceSessionAndToken(inferencesession=session, token="dummy-token")

    def ready(self, id: str) -> bool:
        """Returns True if the session is ready to receive inference requests.

        :param id: The session ID
        :return: True if the session is ready to receive inference requests.
        """
        try:
            session = self.get(id)
            if session is None:
                raise HttpResponseError(f"inferencesessions/{id}", status_code=404)
            return session.status == EntityStatus.admitted
        except:
            return False

    def token(self, id: str, expires: Optional[datetime] = None) -> str:
        """Get an auth token allowing inference calls to the session.

        The local platform does not implement authorization, so this will be a "dummy"
        token.
        """
        return "dummy-token"

    def infer(
        self, id: str, endpoint: str, request: dict[str, Any]
    ) -> list[dict[str, Any]]:
        """Make an inference call to a mock-up inference session.

        :param id: The session ID
        :param endpoint: The endpoint to call
        :param request: The input data for inference
        :return: The inference output
        """
        session = self.get(id)
        if session is None:
            raise HttpResponseError(f"inferencesessions/{id}", status_code=404)
        mock: mocks.InferenceServiceMock = dynamic_import.instantiate(
            session.inferenceService.name
        )
        return mock.infer(endpoint, request)

    def client(
        self,
        session_id: str,
        token: str,
        *,
        interface: Optional[InferenceInterface] = None,
        endpoint: Optional[str] = None,
        input_adapter: Optional[Adapter] = None,
        output_adapter: Optional[Adapter] = None,
    ) -> mocks.InferenceSessionClientMock:
        """Create a mock-up inference session client.

        The token should be one returned either from
        ``inferencesessions.create()`` or from
        ``inferencesessions.token(session_id)``.

        The inference endpoint in the session must also be specified, either
        directly through the ``endpoint`` argument or by specifying an
        ``interface``. Specifying ``interface`` will also use the input and
        output adapters from the interface. You can also specify these
        separately in the ``input_adapter`` and ``output_adapter``. The
        non-``interface`` arguments override the corresponding values in
        ``interface`` if both are specified.

        :param session_id: The inference session to connect to
        :param token: An access token with permission to run inference against
            the session. The token is ignored because the local platform does
            not implement authorization.
        :param interface: The interface to the session. Either ``interface``
            or ``endpoint`` must be specified.
        :param endpoint: The inference endpoint in the session to call. Either
            ``endpoint`` or ``interface`` must be specified.
        :param input_adapter: Optional input adapter, applied to the input
            before sending it to the session. Will override the input adapter
            from ``interface`` if both are specified.
        :param output_adapter: Optional output adapter, applied to the output
            of the session before returning to the client. Will override the
            output adapter from ``interface`` if both are specified.
        :return: An mock-up inference client that makes inference calls to
            the specified session.
        """
        if interface is not None:
            inference_endpoint = endpoint or interface.endpoint
            if input_adapter is None:
                if interface.inputPipeline is not None:
                    input_adapter = create_pipeline(interface.inputPipeline)
            if output_adapter is None:
                if interface.outputPipeline is not None:
                    output_adapter = create_pipeline(interface.outputPipeline)

        session = self.get(session_id)
        if session is None:
            raise HttpResponseError(f"inferencesessions/{session_id}", status_code=404)
        service: mocks.InferenceServiceMock = dynamic_import.instantiate(
            session.inferenceService.name
        )
        return mocks.InferenceSessionClientMock(
            service,
            inference_endpoint=inference_endpoint,
            input_adapter=input_adapter,
            output_adapter=output_adapter,
        )


class _Measurements(_Common[Measurement, MeasurementsOperations]):
    def __init__(self, platform: DyffLocalPlatform):
        super().__init__(
            platform,
            entity_kind=Entities.Measurement,
            entity_type=Measurement,
            remote_ops=platform.remote.measurements if platform.remote else None,
        )

    def create(self, request: SchemaObject[AnalysisCreateRequest]) -> Measurement:
        """Create a new entity.

        :param request: The entity create request specification.
        :return: A full entity spec with its .id and other system properties set
        """
        request = _parse_schema_object(AnalysisCreateRequest, request)
        id = ids.generate_entity_id()

        method_id = request.method
        method = self._platform.methods.get(method_id)
        if method is None:
            raise HttpResponseError(f"methods/{method_id}", status_code=404)

        # Create an entity representing the output of the Analysis
        measurement_spec = method.output.measurement
        if measurement_spec is None:
            raise ValueError("Method spec violates constraints")

        measurement = Measurement(
            account=request.account,
            id=id,
            creationTime=datetime.now(timezone.utc),
            status=EntityStatus.admitted,
            scope=request.scope,
            method=upcast(ForeignMethod, method.dict()),
            arguments=request.arguments,
            inputs=request.inputs,
            **measurement_spec.dict(),
        )
        self._platform._commit(measurement)

        analysis = upcast(Analysis, measurement.dict())
        # This fixes some things that YAML can't parse (like Enums)
        analysis_dict = json.loads(analysis.json())
        analysis_dict["id"] = id
        # Code expects a full k8s manifest
        # TODO: Should only pass the 'spec' part in the first place from
        # dyff-operator, but that requires figuring out how to manipulate YAML in Go
        analysis_dict = {"spec": analysis_dict}
        config_file = self._platform.entity_path(measurement.id) / ".analysis.yaml"
        yaml = ruamel.yaml.YAML()
        with open(config_file, "w") as fout:
            yaml.dump(analysis_dict, fout)

        try:
            run_analysis(
                method,
                storage_root=self._platform.storage_root,
                config_file=config_file,
            )
            measurement.status = EntityStatus.complete
        except Exception:
            measurement.status = EntityStatus.failed
        self._platform._commit(measurement)

        return measurement


class _Methods(_Common[Method, MethodsOperations]):
    def __init__(self, platform: DyffLocalPlatform):
        super().__init__(
            platform,
            entity_kind=Entities.Method,
            entity_type=Method,
            remote_ops=platform.remote.methods if platform.remote else None,
        )

    def create(self, request: SchemaObject[MethodCreateRequest]) -> Method:
        """Create a new entity.

        :param request: The entity create request specification.
        :return: A full entity spec with its .id and other system properties set
        """
        request = _parse_schema_object(MethodCreateRequest, request)
        id = ids.generate_entity_id()
        method_dict = request.dict()
        method_dict["id"] = id
        method_dict["account"] = self._platform.account
        method_dict["status"] = EntityStatus.ready
        method_dict["creationTime"] = datetime.now(timezone.utc)
        method = Method.parse_obj(method_dict)

        self._platform._commit(method)
        return method


class _Models(_Common[Model, ModelsOperations]):
    def __init__(self, platform: DyffLocalPlatform):
        super().__init__(
            platform,
            entity_kind=Entities.Model,
            entity_type=Model,
            remote_ops=platform.remote.models if platform.remote else None,
        )

    def create(self, request: SchemaObject[ModelCreateRequest]) -> Model:
        """Create a new entity.

        :param request: The entity create request specification.
        :return: A full entity spec with its .id and other system properties set
        """
        request = _parse_schema_object(ModelCreateRequest, request)
        id = ids.generate_entity_id()
        model_dict = request.dict()
        model_dict["id"] = id
        model_dict["account"] = self._platform.account
        model_dict["status"] = EntityStatus.ready
        model_dict["creationTime"] = datetime.now(timezone.utc)
        model = Model.parse_obj(model_dict)

        self._platform._commit(model)
        return model


class _Modules(_Common[Module, ModulesOperations]):
    def __init__(self, platform: DyffLocalPlatform):
        super().__init__(
            platform,
            entity_kind=Entities.Module,
            entity_type=Module,
            remote_ops=platform.remote.modules if platform.remote else None,
        )

    def create(self, request: SchemaObject[ModuleCreateRequest]) -> Module:
        """Create a new entity.

        :param request: The entity create request specification.
        :return: A full entity spec with its .id and other system properties set
        """
        request = _parse_schema_object(ModuleCreateRequest, request)
        id = ids.generate_entity_id()
        module_dict = request.dict()
        module_dict["id"] = id
        module_dict["account"] = self._platform.account
        module_dict["status"] = EntityStatus.ready
        module_dict["creationTime"] = datetime.now(timezone.utc)
        module = Module.parse_obj(module_dict)

        cache_dir = self._platform.entity_path(id)
        cache_dir.mkdir()
        with open(cache_dir / ".module.json", "w") as fout:
            fout.write(module.json())
        return module

    def create_package(
        self, package_directory: str, *, account: str, name: str
    ) -> Module:
        """Create a Module resource describing a package structured as a directory tree.

        Internally, constructs a ``ModuleCreateRequest`` using information
        obtained from the directory tree, then calls ``create()`` with the
        constructed request.

        Typical usage::

            module = client.modules.create_package(package_directory, ...)
            client.modules.upload_package(module, package_directory)

        :param package_directory: The root directory of the package.
        :type package_directory: str
        :keyword account: The account that will own the Module resource.
        :type account: str
        :keyword name: The name of the Module resource.
        :type name: str
        :returns: The complete Module resource.
        :rtype: Module
        """
        package_root = Path(package_directory)
        file_paths = [path for path in package_root.rglob("*") if path.is_file()]
        if not file_paths:
            raise ValueError(f"package_directory is empty: {package_directory}")
        artifact_paths = [
            str(Path(file_path).relative_to(package_root)) for file_path in file_paths
        ]
        artifacts = [
            Artifact(
                # FIXME: Is this a useful thing to do? It's redundant with
                # information in 'path'. Maybe it should just be 'code' or
                # something generic.
                kind="".join(file_path.suffixes),
                path=artifact_path,
                digest=Digest(
                    md5=binary.encode(binary.file_digest("md5", str(file_path))),
                ),
            )
            for file_path, artifact_path in zip(file_paths, artifact_paths)
        ]
        request = ModuleCreateRequest(
            account=account,
            name=name,
            artifacts=artifacts,
        )
        return self.create(request)

    def upload_package(self, module: Module, package_directory: str) -> None:
        """Uploads the files in a package directory for which a Module resource has
        already been created.

        Typical usage::

            module = client.modules.create_package(package_directory, ...)
            client.modules.upload_package(module, package_directory)

        :param module: The Module resource for the package.
        :type module: Module
        :param package_directory: The root directory of the package.
        :type package_directory: str
        """
        if any(artifact.digest.md5 is None for artifact in module.artifacts):
            raise ValueError("artifact.digest.md5 must be set for all artifacts")
        for artifact in module.artifacts:
            assert artifact.digest.md5 is not None
            file_path = Path(package_directory) / artifact.path
            cache_path = self._platform.storage_root / module.id / artifact.path
            cache_path.parent.mkdir(parents=True, exist_ok=True)
            with open(file_path, "rb") as fin:
                with open(cache_path, "wb") as fout:
                    fout.write(fin.read())


class _Reports(_Common[Report, ReportsOperations]):
    def __init__(self, platform: DyffLocalPlatform):
        super().__init__(
            platform,
            entity_kind=Entities.Report,
            entity_type=Report,
            remote_ops=platform.remote.reports if platform.remote else None,
        )

    def create(self, request: SchemaObject[ReportCreateRequest]) -> Report:
        """Create a new entity.

        :param request: The entity create request specification.
        :return: A full entity spec with its .id and other system properties set
        """
        request = _parse_schema_object(ReportCreateRequest, request)
        id = ids.generate_entity_id()
        report_dict = request.dict()
        report_dict["id"] = id
        report_dict["account"] = self._platform.account
        report_dict["status"] = EntityStatus.admitted
        report_dict["creationTime"] = datetime.now(timezone.utc)

        evaluation = self._platform.evaluations.get(request.evaluation)
        if evaluation is None:
            raise HttpResponseError(
                f"evaluations/{request.evaluation}", status_code=404
            )
        report_dict["dataset"] = evaluation.dataset
        report_dict["inferenceService"] = (
            evaluation.inferenceSession.inferenceService.id
        )

        report = Report.parse_obj(report_dict)
        self._platform._commit(report)

        try:
            run_report(report, storage_root=self._platform.storage_root)
            report.status = EntityStatus.complete
        except Exception:
            report.status = EntityStatus.failed
        self._platform._commit(report)

        return report


class _SafetyCases(_Common[SafetyCase, SafetycasesOperations]):
    def __init__(self, platform: DyffLocalPlatform):
        super().__init__(
            platform,
            entity_kind=Entities.SafetyCase,
            entity_type=SafetyCase,
            remote_ops=platform.remote.safetycases if platform.remote else None,
        )

    def create(self, request: SchemaObject[AnalysisCreateRequest]) -> SafetyCase:
        """Create a new entity.

        :param request: The entity create request specification.
        :return: A full entity spec with its .id and other system properties set
        """
        request = _parse_schema_object(AnalysisCreateRequest, request)
        id = ids.generate_entity_id()

        method_id = request.method
        method = self._platform.methods.get(method_id)
        if method is None:
            raise HttpResponseError(f"methods/{method_id}", status_code=404)

        # Create an entity representing the output of the Analysis
        safetycase_spec = method.output.safetyCase
        if safetycase_spec is None:
            raise ValueError("Method spec violates constraints")

        # Populate additional context data
        def get_system() -> InferenceService | Model:
            if request.scope.inferenceService is None:
                raise ValueError("Must specify at least request.scope.inferenceService")
            service = self._platform.inferenceservices.get(
                request.scope.inferenceService
            )
            if service is None:
                raise HttpResponseError(
                    f"inferenceservices/{request.scope.inferenceService}",
                    status_code=404,
                )
            if service.model is not None:
                model = self._platform.models.get(service.model.id)
                if model is None:
                    raise HttpResponseError(
                        f"models/{service.model.id}",
                        status_code=404,
                    )
                return model
            else:
                return service

        system = get_system()

        # TODO: Placeholders
        system_documentation = Documentation(
            title=_namespaced_id(system),
            summary="This is the summary text for System Placeholder.",
            entity=system.id,
        )
        usecase_documentation = Documentation(
            title=_namespaced_id(method),
            summary="This is the summary text for Use Case Placeholder.",
            entity=method.id,
        )

        class SystemData(DyffSchemaBaseModel):
            spec: Union[Model, InferenceService]
            documentation: Documentation

        class UseCaseData(DyffSchemaBaseModel):
            spec: Method
            documentation: Documentation

        system_data = SystemData(spec=system, documentation=system_documentation)
        usecase_data = UseCaseData(spec=method, documentation=usecase_documentation)

        def encode(data: DyffSchemaBaseModel) -> str:
            return base64.b64encode(data.json().encode()).decode()

        safetycase = SafetyCase(
            account=request.account,
            id=id,
            creationTime=datetime.now(timezone.utc),
            status=EntityStatus.admitted,
            scope=request.scope,
            method=upcast(ForeignMethod, method.dict()),
            arguments=request.arguments,
            inputs=request.inputs,
            data=[
                AnalysisData(
                    key="system",
                    value=encode(system_data),
                ),
                AnalysisData(
                    key="usecase",
                    value=encode(usecase_data),
                ),
            ],
            # This covers: .name, .description
            **safetycase_spec.dict(),
        )
        self._platform._commit(safetycase)

        analysis = upcast(Analysis, safetycase.dict())
        # This fixes some things that YAML can't parse (like Enums)
        analysis_dict = json.loads(analysis.json())
        analysis_dict["id"] = id
        # Code expects a full k8s manifest
        # TODO: Should only pass the 'spec' part in the first place from
        # dyff-operator, but that requires figuring out how to manipulate YAML in Go
        analysis_dict = {"spec": analysis_dict}
        config_file = self._platform.entity_path(safetycase.id) / ".analysis.yaml"
        yaml = ruamel.yaml.YAML()
        with open(config_file, "w") as fout:
            yaml.dump(analysis_dict, fout)

        # This mimics the platform behavior of re-trying failed analyses in
        # case the error was due to a transient cloud platform issue. We don't
        # expect errors to be transient when running locally, but it's necessary
        # for testing certain regressions (e.g., [DYFF-577]).
        # FIXME: Magic number
        for _retry in range(2):
            try:
                run_analysis(
                    method,
                    storage_root=self._platform.storage_root,
                    config_file=config_file,
                )
                safetycase.status = EntityStatus.complete
                break
            except Exception:
                safetycase.status = EntityStatus.failed
        self._platform._commit(safetycase)

        # Populate the .id field of the scores
        try:
            file = self._platform.entity_path(safetycase.id) / ".dyff" / "scores.json"
            with open(file, "r") as fin:
                scores_obj = json.load(fin)
        except FileNotFoundError:
            pass
        else:
            for score_data in scores_obj["scores"]:
                score_data["kind"] = Entities.Score.value
                score_data["id"] = ids.namespaced_id(
                    score_data["analysis"], score_data["name"]
                )
            with open(file, "w") as fout:
                json.dump(scores_obj, fout)

        return safetycase


class _Scores:
    def __init__(self, platform: DyffLocalPlatform):
        self._platform = platform

    def get(self, *, analysis: str) -> list[Score]:
        """Get all the scores output by the given analysis."""
        try:
            file = self._platform.entity_path(analysis) / ".dyff" / "scores.json"
            with open(file, "r") as fin:
                scores_obj = json.load(fin)
            return [Score.parse_obj(score) for score in scores_obj["scores"]]
        except FileNotFoundError:
            return []


class DyffLocalPlatform:
    """Emulates a subset of Dyff Platform operations locally.

    This class is intended to aid in local development and debugging of code
    and data that will run on the Dyff Platform. The inferface mirrors that of
    ``dyff.client.Client`` and should have similar behavior.

    Entities created on the local platform are stored in a local cache. You can
    optionally provide a ``dyff.client.Client`` instance, in which case you can
    use the ``.fetch()`` functions to download an entity and all associated
    artifacts to local storage.
    """

    def __init__(
        self,
        storage_root: Path | str = "dyff-outputs",
        *,
        remote_client: Optional[Client] = None,
    ):
        self._storage_root = Path(storage_root).resolve()
        self._client = remote_client

        self._datasets = _Datasets(self)
        self._evaluations = _Evaluations(self)
        self._inferenceservices = _InferenceServices(self)
        self._inferencesessions = _InferenceSessions(self)
        self._measurements = _Measurements(self)
        self._methods = _Methods(self)
        self._models = _Models(self)
        self._modules = _Modules(self)
        self._reports = _Reports(self)
        self._safetycases = _SafetyCases(self)
        self._scores = _Scores(self)

        self._storage_root.mkdir(exist_ok=True)

    def _require_client(self) -> Client:
        if not self._client:
            raise ValueError("remote client not available")
        return self._client

    @property
    def datasets(self) -> _Datasets:
        """Operations on :class:`~dyff.schema.platform.Dataset` entities."""
        return self._datasets

    @property
    def evaluations(self) -> _Evaluations:
        """Operations on :class:`~dyff.schema.platform.Evaluation` entities."""
        return self._evaluations

    @property
    def inferenceservices(self) -> _InferenceServices:
        """Operations on :class:`~dyff.schema.platform.InferenceService` entities."""
        return self._inferenceservices

    @property
    def inferencesessions(self) -> _InferenceSessions:
        """Operations on :class:`~dyff.schema.platform.InferenceSession` entities."""
        return self._inferencesessions

    @property
    def measurements(self) -> _Measurements:
        """Operations on :class:`~dyff.schema.platform.Measurement` entities."""
        return self._measurements

    @property
    def methods(self) -> _Methods:
        """Operations on :class:`~dyff.schema.platform.Method` entities."""
        return self._methods

    @property
    def models(self) -> _Models:
        """Operations on :class:`~dyff.schema.platform.Model` entities."""
        return self._models

    @property
    def modules(self) -> _Modules:
        """Operations on :class:`~dyff.schema.platform.Module` entities."""
        return self._modules

    @property
    def reports(self) -> _Reports:
        """Operations on :class:`~dyff.schema.platform.Report` entities."""
        return self._reports

    @property
    def safetycases(self) -> _SafetyCases:
        """Operations on :class:`~dyff.schema.platform.SafetyCase` entities."""
        return self._safetycases

    @property
    def scores(self) -> _Scores:
        """Operations on :class:`~dyff.schema.platform.Score` entities."""
        return self._scores

    @property
    def remote(self) -> Client | None:
        """The remote client instance."""
        return self._client

    @property
    def storage_root(self) -> Path:
        """The root of the local storage directory tree."""
        return self._storage_root

    @property
    def account(self) -> str:
        """A dummy local account name."""
        return "local"

    def entity_path(self, id: str) -> Path:
        """Returns the path to the local directory that stores information about the
        entity with the given ID."""
        return self.storage_root / id

    def link_entity(self, package: Path, id: Optional[str] = None) -> str:
        """Creates a symlink in the local storage tree pointing to a directory on the
        file system. You can either provide an ID or let the platform generate one.
        Returns the assigned ID.

        This can be used to create an "editable" package so that you can refer to the
        package at a stable ID while still changing the contents of the files in it.

        :param package: The path to the directory to symlink to.
        :param id: If provided, becomes the ID of the linked package in the platform.
            Otherwise, a new ID is generated.
        :return: The ID of the linked entity.
        """
        package = package.resolve()
        id = id or ids.generate_entity_id()
        cache_path = self.storage_root / id
        if not cache_path.exists():
            cache_path.symlink_to(package)
        elif cache_path.is_symlink():
            target = cache_path.resolve()
            if target != package:
                raise ValueError(f"{id} -> {target}: conflict")
        else:
            raise ValueError(f"{cache_path} exists and is not a symlink")
        return id

    def _commit(self, entity: DyffEntity) -> None:
        cache_dir = self.entity_path(entity.id)
        cache_dir.mkdir(exist_ok=True)
        filename = f".{entity.kind.lower()}.json"
        with open(cache_dir / filename, "w") as fout:
            fout.write(entity.json())



# SPDX-FileCopyrightText: 2024 UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

import errno
import os
import re
import subprocess
import sys
from pathlib import Path
from typing import Optional

from dyff.schema.platform import MethodBase, MethodImplementationKind, Report

from .._internal import fqn
from . import context, jupyter, legacy, python


# https://stackoverflow.com/a/38662876/3709935
def _remove_ansi_escape_sequences(line):
    ansi_escape = re.compile(r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])")
    return ansi_escape.sub("", line)


def _delete_directory_contents(
    directory: Path, *, keep: Optional[list[Path]] = None
) -> None:
    if keep is None:
        keep = []
    if not directory.is_dir():
        raise ValueError(f"path must be a directory: {directory}")
    for p in directory.iterdir():
        if p.is_dir():
            _delete_directory_contents(p, keep=keep)
            try:
                p.rmdir()
            except OSError as ex:
                # Directory might not be empty if a sub-path is in 'keep'
                if ex.errno != errno.ENOTEMPTY:
                    raise
        elif all(not k.exists() or not p.samefile(k) for k in keep):
            p.unlink()


def _log_separator() -> str:
    lines = [
        "[dyff] " + ("=" * 72),
        "[dyff] START OF ANALYSIS RUN",
        "[dyff] " + ("=" * 72) + "\n",
    ]
    return "\n".join(lines)


def run_analysis(method: MethodBase, *, storage_root: Path, config_file: Path) -> None:
    """Run an analysis workflow locally.

    The analysis workflow consists of running a Method on specified inputs to
    generate output artifacts such as SafetyCases and Measurements.

    The analysis is run in a sub-process. Output artifacts will be created in
    ``storage_root / analysis_id``. Logs will be saved in
    ``storage_root / analysis_id / .dyff / logs.txt``.

    :param method: The specification of the method to run.
    :keyword storage_root: The root directory for local storage of input and
        output artifacts.
    :keyword config_file: A YAML file containing a Kubernetes Analysis resource
        (kind ``analyses.dyff.io/v1alpha1``).

        .. note::

            The format of this file is expected to change in a future release.
    """
    # Need this to get the ID assigned to the analysis
    analysis_id = context.id_from_config_file(config_file)
    output_dir = storage_root / analysis_id

    pythonpath = os.pathsep.join(
        str(storage_root / module) for module in method.modules
    )
    env = os.environ.copy()
    env.update(
        {
            "DYFF_AUDIT_LOCAL_STORAGE_ROOT": str(storage_root),
            "DYFF_AUDIT_ANALYSIS_CONFIG_FILE": str(config_file),
            "PYTHONPATH": pythonpath,
        }
    )

    if method.implementation.kind == MethodImplementationKind.JupyterNotebook:
        impl_module, impl_name = fqn(jupyter.run_jupyter_notebook)
    elif method.implementation.kind == MethodImplementationKind.PythonFunction:
        impl_module, impl_name = fqn(python.run_python_function)
    elif method.implementation.kind == MethodImplementationKind.PythonRubric:
        impl_module, impl_name = fqn(python.run_python_rubric)
    else:
        raise NotImplementedError(
            f"method.implementation.kind = {method.implementation.kind}"
        )

    log_file = output_dir / ".dyff" / "logs.txt"
    log_file.parent.mkdir(parents=True, exist_ok=True)
    # Make sure log_file exists so we don't delete its parent directory
    log_file.touch()

    # [DYFF-577] If the analysis is re-tried due to an error, there may be
    # partial output already from the previous run. This can cause additional
    # spurious errors (e.g., for outputting "duplicate" data) that mask the
    # original error. Note that the config file may be in the output_dir tree
    # and must be kept if present. We also keep the log file because we're
    # appending each run to it.
    _delete_directory_contents(output_dir, keep=[config_file, log_file])

    try:
        with open(log_file, "ab", buffering=0) as fout:
            fout.write(_log_separator().encode())
            cmd = f"from {impl_module} import {impl_name}; {impl_name}()"
            subprocess.run(
                ["python3", "-u", "-X", "faulthandler", "-c", cmd],
                env=env,
                check=True,
                # Redirect both streams to log file
                stdout=fout,
                stderr=subprocess.STDOUT,
            )
    finally:
        # [DYFF-509] Output from Jupyter notebooks often has ANSI escape
        # sequences in it. We want to remove these sequences, but without
        # shadowing the original exception (if any) and without leaving a moment
        # where the logs.txt file could be lost if there's another error.
        current_exception = sys.exc_info()[1]
        try:
            scratch_file = Path(str(log_file) + ".tmp")
            with open(log_file, "r") as fin:
                with open(scratch_file, "w") as fout:
                    fout.writelines(_remove_ansi_escape_sequences(line) for line in fin)
            scratch_file.rename(log_file)
        except Exception as ex:
            if not current_exception:
                raise ex


def run_report(report: Report, *, storage_root: Path):
    """Run a Report workflow locally.

    .. deprecated:: 0.8.0

        Report functionality has been refactored into the
        Method/Measurement/Analysis apparatus. Creation of new Reports is
        disabled.

    The workflow is run in a sub-process. Output artifacts will be created in
    ``storage_root / report_id``. Logs will be saved in
    ``storage_root / report_id / .dyff / logs.txt``.

    :param report: The specification of the Report to run.
    :keyword storage_root: The root directory for local storage of input and
        output artifacts.
    """
    return legacy_run_report(
        rubric=report.rubric,
        dataset_path=str(storage_root / report.dataset),
        evaluation_path=str(storage_root / report.evaluation),
        output_path=str(storage_root / report.id),
        modules=[str(storage_root / module) for module in report.modules],
    )


def legacy_run_report(
    *,
    rubric: str,
    dataset_path: str,
    evaluation_path: str,
    output_path: str,
    modules: Optional[list[str]] = None,
):
    if modules is None:
        modules = []

    def quote(s) -> str:
        return f'"{s}"'

    args = [
        quote(rubric),
        quote(dataset_path),
        quote(evaluation_path),
        quote(output_path),
        ", ".join(quote(module) for module in modules),
    ]

    impl_module, impl_name = fqn(legacy.run_python_rubric)
    cmd = (
        f"from {impl_module} import {impl_name}; {impl_name}"
        "(rubric={}, dataset_path={}, evaluation_path={}, output_path={}, modules=[{}])".format(
            *args
        )
    )

    pythonpath = os.pathsep.join(str(module) for module in modules)
    env = os.environ.copy()
    env.update({"PYTHONPATH": pythonpath})

    log_file = Path(output_path) / ".dyff" / "logs.txt"
    log_file.parent.mkdir(parents=True, exist_ok=True)
    with open(log_file, "wb", buffering=0) as fout:
        subprocess.run(
            ["python3", "-u", "-X", "faulthandler", "-c", cmd],
            env=env,
            check=True,
            # Redirect both streams to log file
            stdout=fout,
            stderr=subprocess.STDOUT,
        )


__all__ = [
    "legacy_run_report",
    "run_analysis",
    "run_report",
]



# SPDX-FileCopyrightText: 2024 UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

from typing import Any, TypeVar

from dyff.schema.base import DyffBaseModel


def fqn(obj) -> tuple[str, str]:
    """See: https://stackoverflow.com/a/70693158"""
    try:
        module = obj.__module__
    except AttributeError:
        module = obj.__class__.__module__
    try:
        name = obj.__qualname__
    except AttributeError:
        name = obj.__class__.__qualname__
    # if obj is a method of builtin class, then module will be None
    if module == "builtins" or module is None:
        raise AssertionError("should not be called on a builtin")
    return module, name


_DyffModelT = TypeVar("_DyffModelT", bound=DyffBaseModel)


def upcast(t: type[_DyffModelT], obj: dict[str, Any]) -> _DyffModelT:
    fields = {k: v for k, v in obj.items() if k in t.__fields__}
    return t.parse_obj(fields)


__all__ = [
    "fqn",
    "upcast",
]



# SPDX-FileCopyrightText: 2024 UL Research Institutes
# SPDX-License-Identifier: Apache-2.0
"""Base class for Rubrics."""

# mypy: disable-error-code="import-untyped"
from __future__ import annotations

import abc
from typing import Iterable

import pyarrow


class Rubric(abc.ABC):
    """A Rubric is a mechanism for producing inference-level performance scores for the
    output of an inference service on a given inference task."""

    @property
    @abc.abstractmethod
    def name(self) -> str:
        """The "semi-qualified" type name of the Rubric.

        The result should be
        such that ``f"alignmentlabs.audit.scoring.{self.name}"`` is the
        fully-qualified name of the type.
        """
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def schema(self) -> pyarrow.Schema:
        """The PyArrow schema of the output of applying the Rubric."""
        raise NotImplementedError()

    @abc.abstractmethod
    def apply(
        self, task_data: pyarrow.datasets.Dataset, predictions: pyarrow.datasets.Dataset
    ) -> Iterable[pyarrow.RecordBatch]:
        """Create a generator that yields batches of scored inferences."""
        raise NotImplementedError()



# SPDX-FileCopyrightText: 2024 UL Research Institutes
# SPDX-License-Identifier: Apache-2.0
"""Tools for working with input and output data for systems being audited."""

from . import text

__all__ = [
    "text",
]



# SPDX-FileCopyrightText: 2024 UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

# mypy: disable-error-code="import-untyped"
from __future__ import annotations

import base64
import json
import os
from pathlib import Path
from typing import Any, Literal, Optional, Union

import pyarrow
import pydantic
import ruamel.yaml
from IPython.display import display as ipy_display

from dyff.schema import ids
from dyff.schema.dataset import arrow
from dyff.schema.platform import (
    Analysis,
    Documentation,
    DyffSchemaBaseModel,
    InferenceService,
    Method,
    Model,
    ScoreData,
    ScoreMetadata,
    ScoreMetadataRefs,
)

from .. import components
from .._internal import timestamp, upcast


def _decode_data(data: str) -> Any:
    return json.loads(base64.b64decode(data))


def _analysis_from_yaml(analysis_yaml: dict) -> Analysis:
    if "spec" in analysis_yaml:
        analysis_yaml = analysis_yaml["spec"]

    if (analysis_data := analysis_yaml.get("data")) is not None:
        for item in analysis_data:
            if item["key"] == "analysis":
                return upcast(Analysis, _decode_data(item["value"]))

    analysis_yaml = analysis_yaml.copy()
    analysis_yaml["method"].setdefault("id", ids.null_id())
    analysis_yaml["method"].setdefault("account", ids.null_id())
    return upcast(Analysis, analysis_yaml)


def _id_from_yaml(analysis_yaml: dict) -> str:
    return analysis_yaml["spec"]["id"]


def id_from_config_file(analysis_config_file: Union[Path, str]) -> str:
    """Parses an analysis config file and returns the analysis ID."""
    yaml = ruamel.yaml.YAML()
    with open(analysis_config_file, "r") as fin:
        analysis_yaml = yaml.load(fin)
    return _id_from_yaml(analysis_yaml)


class SystemInformation(DyffSchemaBaseModel):
    spec: Union[Model, InferenceService] = pydantic.Field(
        description="The specification of the system entity. This is a Model"
        " if the system is backed by a model, otherwise an InferenceService."
    )
    documentation: Documentation = pydantic.Field(
        description="The documentation associated with the system entity."
    )


class UseCaseInformation(DyffSchemaBaseModel):
    spec: Method = pydantic.Field(
        description="The specification of the use case entity."
    )
    documentation: Documentation = pydantic.Field(
        description="The documentation associated with the system entity."
    )


class Scores(DyffSchemaBaseModel):
    scores: list[ScoreData] = pydantic.Field(
        default_factory=list, description="The list of score data"
    )


class AnalysisContext:
    """AnalysisContext is Dyff's mechanism for making input data available to user-
    authored analysis Methods.

    When the Method is implemented in a framework such as Jupyter that does not support
    "arguments", the implementation accesses its inputs by instantiating an
    AnalysisContext. The AnalysisContext gets its configuration information from
    environment variables. The runners for analyses implemented in other ways also use
    AnalysisContext under the hood.
    """

    def __init__(
        self,
        *,
        analysis_config_file: Union[Path, str, None] = None,
        local_storage_root: Union[Path, str, None] = None,
        analysis: Optional[Analysis] = None,
        id: Optional[str] = None,
        allow_override_from_environment: bool = False,
    ):
        """When running an analysis on the Dyff Platform, the platform provides the
        ``analysis_config_file`` and the ``local_storage_root`` arguments via the
        environment variables ``DYFF_AUDIT_ANALYSIS_CONFIG_FILE`` and
        ``DYFF_AUDIT_LOCAL_STORAGE_ROOT``.

        .. note::

            If you are creating an ``AnalysisContext`` instance in code that
            will run on the Dyff Platform, you must call the constructor
            with **no arguments**, e.g., ``ctx = AnalysisContext()``.

        :param analysis_config_file: The path to a YAML-format specification of
            an Analysis. If not specified, it is read from the
            ``DYFF_AUDIT_ANALYSIS_CONFIG_FILE`` environment variable.
        :param local_storage_root: The root directory for local storage of
            entity data. If not specified, it is read from the
            ``DYFF_AUDIT_LOCAL_STORAGE_ROOT`` environment variable.
        :param analysis: You can also specify the analysis as an Analysis
            instance. If you do, you must also specify the ``id``. This is mainly
            useful for debugging.
        :param id: The ID of the analysis, which is needed when instantiating
            from an Analysis instance, because Analysis doesn't have an ``.id``
            field.
        :param allow_override_from_environment: If ``True``, environment
            variables will override values in the config file. By default, the
            config file has precedence.
        """
        if id is not None and analysis is not None:
            if analysis_config_file is not None:
                raise ValueError(
                    "'(id, analysis)' and 'analysis_config_file' are mutually exclusive"
                )
            self._id = id
            self._analysis = analysis
        else:
            if allow_override_from_environment:
                analysis_config_file = (
                    os.environ.get("DYFF_AUDIT_ANALYSIS_CONFIG_FILE")
                    or analysis_config_file
                )
            else:
                analysis_config_file = analysis_config_file or os.environ.get(
                    "DYFF_AUDIT_ANALYSIS_CONFIG_FILE"
                )
            if analysis_config_file is None:
                raise ValueError(
                    "Must provide '(id, analysis)' or 'analysis_config_file'"
                    " or set DYFF_AUDIT_ANALYSIS_CONFIG_FILE environment variable"
                )
            if id is not None or analysis is not None:
                raise ValueError(
                    "'(id, analysis)' and 'analysis_config_file' are mutually exclusive"
                )

            yaml = ruamel.yaml.YAML()
            with open(analysis_config_file, "r") as fin:
                analysis_yaml = yaml.load(fin)
            self._id = _id_from_yaml(analysis_yaml)
            self._analysis = _analysis_from_yaml(analysis_yaml)

        if allow_override_from_environment:
            local_storage_root = (
                os.environ.get("DYFF_AUDIT_LOCAL_STORAGE_ROOT") or local_storage_root
            )
        else:
            local_storage_root = local_storage_root or os.environ.get(
                "DYFF_AUDIT_LOCAL_STORAGE_ROOT"
            )
        if local_storage_root is None:
            raise ValueError(
                "Must provide local_storage_root"
                " or set DYFF_AUDIT_LOCAL_STORAGE_ROOT environment variable."
            )
        self._local_storage_root = Path(local_storage_root)
        if not self._local_storage_root.is_absolute():
            raise ValueError("local_storage_root must be an absolute path")

        self._parameters = {p.keyword: p for p in self.analysis.method.parameters}
        self._arguments = {a.keyword: a.value for a in self.analysis.arguments}
        self._inputs = {i.keyword: i.entity for i in self.analysis.inputs}
        self._input_kinds = {i.keyword: i.kind for i in self.analysis.method.inputs}
        self._input_paths = {
            e.keyword: str(self._local_storage_root / e.entity)
            for e in self.analysis.inputs
        }

        self._analysis_data = {
            e.key: _decode_data(e.value) for e in self._analysis.data
        }

        system_data = self._analysis_data.get("system")
        self._system_information = (
            SystemInformation(**system_data) if system_data else None
        )

        usecase_data = self._analysis_data.get("usecase")
        self._usecase_information = (
            UseCaseInformation(**usecase_data) if usecase_data else None
        )

    @property
    def id(self) -> str:
        """The ID of the current analysis."""
        return self._id

    @property
    def analysis(self) -> Analysis:
        """The spec for the current analysis."""
        return self._analysis

    @property
    def local_storage_root(self) -> Path:
        """The root path where subdirectories containing artifacts for individual Dyff
        resources will be created on the local file system."""
        return self._local_storage_root

    @property
    def output_path(self) -> Path:
        """The path where output artifacts are stored on the local file system."""
        return self._local_storage_root / self._id

    @property
    def arguments(self) -> dict[str, str]:
        """The arguments passed to the analysis."""
        return self._arguments.copy()

    @property
    def inputs(self) -> list[str]:
        """The names of all of the input datasets passed to the analysis."""
        return list(self._inputs.keys())

    def get_argument(self, keyword: str) -> str:
        """Get the value of an argument passed to the analysis.

        :param keyword: The keyword specified for the argument in the Method spec
        """
        return self._arguments[keyword]

    def open_input_dataset(self, keyword: str) -> pyarrow.dataset.Dataset:
        """Open a dataset provided as input to the analysis.

        :param keyword: The keyword specified for the input in the Method spec
        """
        entity = self._inputs[keyword]
        path = self._local_storage_root / entity
        return arrow.open_dataset(str(path))

    @property
    def system(self) -> Optional[SystemInformation]:
        """Information about the system under test.

        Currently, this is populated only for the SafetyCase workflow.
        """
        return self._system_information

    @property
    def usecase(self) -> Optional[UseCaseInformation]:
        """Information about the use case being tested.

        Currently, this is populated only for the SafetyCase workflow.

        .. deprecated:: 0.10.2

            "usecase" is a deprecated alias for "Method"; this name will
            change in a future release.
        """
        return self._usecase_information

    def Conclusion(
        self,
        *,
        text: str,
        indicator: Literal["Information", "Question", "Hazard"] = "Information",
    ) -> None:
        """Display a :py:class:`~dyff.audit.components.Conclusion` widget at the current
        position in the Jupyter notebook.

        :param text: The text to display.
        :param indicator: The icon to display.
        """
        component = components.Conclusion(indicator=indicator, text=text)
        ipy_display(component)

    def Score(
        self,
        *,
        quantity: float,
        text: str,
        output: Optional[str] = None,
        display: bool = True,
        format: Optional[str] = None,
        unit: Optional[str] = None,
    ) -> None:
        """Display a :py:class:`~dyff.audit.components.Score` widget at the current
        position in the Jupyter notebook.

        If ``output`` is given, the score will be saved in the Dyff datastore
        under the specified name. The name must match the name of a score
        declared in the Method spec. In this case, ``format`` and ``unit`` take
        the value specific in the spec, and overriding this value is an error.

        :param quantity: The measured value of the score.
        :param text: A text description of what the score means.
        :param output: If given, it must match the name of a score declared
            in the Method spec. The score quantity will be saved in the Dyff
            datastore under that name.
        :param display: If False, do not display the score widget in the
            Jupyter notebook (but still output the score if ``output`` is given).
        :param format: A Python format string used to render the quantity as
            a string. It *must* use the key ``quantity``, and it *may* use the
            key ``unit``, e.g. ``"{quantity} {unit}"``.
        :param unit: A string representation of the unit of measurement, e.g.,
            ``"MJ/kg"``, ``"%"``, etc. Prefer SI units when applicable.
        """
        quantity_string: Optional[str] = None

        if output:
            for spec in self.analysis.method.scores:
                if spec.name == output:
                    break
            else:
                raise ValueError(
                    f"No score declared with name '{output}'; see Method {self.analysis.method.id}"
                )
            if format is not None or unit is not None:
                raise ValueError("Cannot override 'format' or 'unit' from score spec")
            format = spec.format
            unit = spec.unit
            quantity_string = spec.quantity_string(quantity)

            score = ScoreData(
                analysis=self.id,
                metadata=ScoreMetadata(
                    refs=ScoreMetadataRefs(
                        method=self.analysis.method.id,
                        **self.analysis.scope.dict(),
                    )
                ),
                quantity=quantity,
                quantityString=quantity_string,
                text=text,
                **spec.dict(),
            )

            meta_dir = self.output_path / ".dyff"
            meta_dir.mkdir(parents=True, exist_ok=True)
            try:
                scores_data: Scores = Scores.parse_file(meta_dir / "scores.json")
            except FileNotFoundError:
                scores_data = Scores()

            for existing_score in scores_data.scores:
                if existing_score.name == output:
                    raise ValueError(f"Already specified: score '{output}'")

            scores_data.scores.append(score)
            with open(meta_dir / "scores.json", "w") as fout:
                fout.write(scores_data.json())

        if display:
            if format is None:
                raise ValueError(
                    "must specifiy 'format' for Scores with no ScoreSpec declared"
                )
            if quantity_string is None:
                quantity_string = ScoreData.format_quantity(format, quantity, unit=unit)
            kwargs = dict(quantity=quantity_string, text=text)
            # We want None to mean "unspecified" so that we get the defaults
            if unit is not None:
                kwargs["unit"] = unit
            component = components.Score.parse_obj(kwargs)
            ipy_display(component)

    def TitleCard(
        self,
        *,
        headline: str,
        author: str,
        summary_phrase: str,
        summary_text: str,
        system_title: str | None = None,
        system_summary: str | None = None,
        usecase_title: str | None = None,
        usecase_summary: str | None = None,
    ) -> None:
        """Display a :py:class:`~dyff.audit.components.TitleCard` widget at the current
        position in the Jupyter notebook.

        Normally, this should be the first output in the notebook.

        :param headline: The headline text for the notebook.
        :param author: Description of the author(s) of the notebook.
        :param summary_phrase: A "sub-heading" for the summary information.
        :param summary_text: A text summary of the notebook.
        :param system_title: The "title" of the system-under-test. When running on
            the Dyff platform, this defaults to the title given in the system's
            documentation.
        :param system_summary: A "summary" of the system-under-test. When running on
            the Dyff platform, this defaults to the summary given in the
            system's documentation.
        :param usecase_title: The "title" of the Method being run. When running on
            the Dyff platform, this defaults to the title given in the Method's
            documentation.

            .. deprecated:: 0.10.2

                "usecase" is a deprecated alias for "Method"; this name will
                change in a future release.

        :param usecase_summary: A "summary" of the Method being run. When running on
            the Dyff platform, this defaults to the summary given in the
            Methods's documentation.

            .. deprecated:: 0.10.2

                "usecase" is a deprecated alias for "Method"; this name will
                change in a future release.
        """

        def from_context(name: str, path: str) -> str:
            keys = path.split(".")
            d = self._analysis_data
            try:
                for k in keys:
                    d = d[k]
                if d is None or not isinstance(d, str):
                    raise ValueError()
                return d
            except Exception:
                raise ValueError(
                    f"Must set {name} because {path} is not present in analysis context."
                )

        if system_title is None:
            system_title = from_context("system_title", "system.documentation.title")
        if system_summary is None:
            system_summary = from_context(
                "system_summary", "system.documentation.summary"
            )
        if usecase_title is None:
            usecase_title = from_context("usecase_title", "usecase.documentation.title")
        if usecase_summary is None:
            usecase_summary = from_context(
                "usecase_summary", "usecase.documentation.summary"
            )

        if (date := self._analysis_data.get("date")) is not None:
            # Validate the date
            date = timestamp.dt_to_str(timestamp.parse(date))
        else:
            date = timestamp.now_str()

        component = components.TitleCard(
            headline=headline,
            author=author,
            date=date,
            system_title=system_title,
            system_summary=system_summary,
            usecase_title=usecase_title,
            usecase_summary=usecase_summary,
            summary_phrase=summary_phrase,
            summary_text=summary_text,
        )
        ipy_display(component)


__all__ = [
    "AnalysisContext",
    "Scores",
    "id_from_config_file",
]



# SPDX-FileCopyrightText: 2024 UL Research Institutes
# SPDX-License-Identifier: Apache-2.0
"""Base class for Metrics."""

# mypy: disable-error-code="import-untyped"
from __future__ import annotations

import abc

import pandas
import pyarrow


class Metric(abc.ABC):
    """A Metric is an operation that can be applied to a set of inference-level scores
    to produce an aggregate summary of performance."""

    @property
    @abc.abstractmethod
    def name(self) -> str:
        """The "semi-qualified" type name of the Metric.

        The result should be
        such that ``f"alignmentlabs.audit.metrics.{self.name}"`` is the
        fully-qualified name of the type.
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def schema(self) -> pyarrow.Schema:
        """The PyArrow schema of the output of applying the Metric."""
        raise NotImplementedError

    @abc.abstractmethod
    def __call__(self, scores: pandas.DataFrame) -> pandas.DataFrame:
        """Compute the metric."""
        raise NotImplementedError



# SPDX-FileCopyrightText: 2024 UL Research Institutes
# SPDX-License-Identifier: Apache-2.0
"""Rubrics related to supervised classification."""

# mypy: disable-error-code="import-untyped"
from __future__ import annotations

import logging
from typing import Any, Iterable, List

import pyarrow
import pydantic

from dyff.schema.dataset import ReplicatedItem
from dyff.schema.dataset.arrow import arrow_schema, schema_function

from .base import Rubric


def top_k(prediction: Any | List[Any], truth: Any, *, k: int) -> int:
    """Return ``1`` if any of the first ``k`` elements of ``prediction`` are equal to
    ``truth``, and ``0`` otherwise.

    :param prediction: Either a list of predicted labels in descending order of
      "score", or a single predicted label. For a single label, all values
      of ``k`` are equivalent to ``k = 1``.
    :type prediction: Any or List(Any)
    :param truth: The true label.
    :type truth: Any
    :param k: The number of predictions to consider.
    :type k: int
    :return: ``1`` if one of the top ``k`` predictions was correct, ``0`` otherwise.
    :rtype: int
    """
    if k < 1:
        raise ValueError("'k' must be >= 1")
    if isinstance(prediction, list):
        last = min(k, len(prediction))
        return int(truth in prediction[:last])
    else:
        return truth == prediction


class TopKAccuracyScoredItem(ReplicatedItem):
    """Placeholder."""

    top1: bool = pydantic.Field(
        ...,
        description="0-1 indicator of whether the top-1 prediction was correct.",
    )
    top5: bool = pydantic.Field(
        ...,
        description="0-1 indicator of whether any of the top-5 predictions were correct.",
    )


class TopKAccuracy(Rubric):
    """Computes top-1 and top-5 accuracy for classification tasks."""

    def __init__(self):
        self._ks = [1, 5]

    def _build_labels_dict(self, labels_dataset):
        labels = {}
        for b in labels_dataset.to_batches(columns=["_index_", "label"]):
            df = b.to_pandas()
            for index, label in zip(df["_index_"], df["label"]):
                labels[index] = label
        return labels

    @property
    def name(self) -> str:
        return "classification.TopKAccuracy"

    @property
    @schema_function(arrow_schema(TopKAccuracyScoredItem))
    def schema(self) -> pyarrow.Schema:
        """The PyArrow schema of the output of applying the Rubric.

        There is one row per input instance. A prediction is scored as correct if
        ``prediction == truth``. For top-5 accuracy, the instance is correct if
        any of the top-5 predictions was correct.
        """

    def apply(
        self, task_data: pyarrow.datasets.Dataset, predictions: pyarrow.datasets.Dataset
    ) -> Iterable[pyarrow.RecordBatch]:
        labels_dict = self._build_labels_dict(task_data)
        processed = []
        for b in predictions.to_batches(
            columns=["_index_", "_replication_", "responses"]
        ):
            batch = []
            for item in b.to_pylist():
                index = item["_index_"]
                replication = item["_replication_"]
                responses = item["responses"]
                truth = labels_dict[index]
                scored_item = TopKAccuracyScoredItem(
                    _index_=index,
                    _replication_=replication,
                    top1=bool(top_k(responses, truth, k=1)),
                    top5=bool(top_k(responses, truth, k=5)),
                )
                batch.append(scored_item.dict())
                processed.append(index)
                if len(processed) % 100000 == 0:
                    logging.info(f"processed {len(processed)} items")
            yield pyarrow.RecordBatch.from_pylist(batch, schema=self.schema)
        logging.info(f"final: {len(processed)} items; {len(set(processed))} unique")


__all__ = [
    "TopKAccuracy",
    "TopKAccuracyScoredItem",
    "top_k",
]



# SPDX-FileCopyrightText: 2024 UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

from .context import AnalysisContext, Scores
from .runners import legacy_run_report, run_analysis, run_report

__all__ = [
    "AnalysisContext",
    "Scores",
    "legacy_run_report",
    "run_analysis",
    "run_report",
]



# SPDX-FileCopyrightText: 2024 UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

# mypy: disable-error-code="import-untyped"
from __future__ import annotations

import importlib.resources
import string
from typing import Literal, Optional

import pydantic

from dyff.schema.platform import DyffSchemaBaseModel, summary_maxlen

from . import templates


class Conclusion(DyffSchemaBaseModel):
    indicator: Literal["Hazard", "Information", "Question"] = pydantic.Field(
        default="Information",
        description="The kind of indicator icon to display in the conclusion.",
    )

    text: str = pydantic.Field(
        description="A short text description of the conclusion.", max_length=140
    )

    def _repr_html_(self) -> str:
        template_file = (
            importlib.resources.files(templates) / "Conclusion.template.html"
        )
        with template_file.open("r") as fin:
            template = string.Template(fin.read())

        return template.substitute(
            text=self.text,
            info_display="block" if self.indicator == "Information" else "none",
            question_display="block" if self.indicator == "Question" else "none",
            hazard_display="block" if self.indicator == "Hazard" else "none",
        )


class Score(DyffSchemaBaseModel):
    quantity: str = pydantic.Field(description="The string rendering of the quantity")

    text: str = pydantic.Field(
        description="A short text interpretation of the quantity.",
        max_length=summary_maxlen(),
    )

    unit: Optional[str] = pydantic.Field(
        default=None,
        description="The unit of measure, if applicable (e.g., 'meters', 'kJ/g')."
        " Use standard SI abbreviations where possible for better indexing.",
    )

    def _repr_html_(self) -> str:
        template_file = importlib.resources.files(templates) / "Score.template.html"
        with template_file.open("r") as fin:
            template = string.Template(fin.read())

        return template.substitute(
            quantity=self.quantity,
            text=self.text,
        )


class TitleCard:
    def __init__(
        self,
        *,
        headline: str,
        author: str,
        date: str,
        system_title: str,
        system_summary: str,
        usecase_title: str,
        usecase_summary: str,
        summary_phrase: str,
        summary_text: str,
    ):
        self._headline = headline
        self._author = author
        self._date = date
        self._system_title = system_title
        self._system_summary = system_summary
        self._usecase_title = usecase_title
        self._usecase_summary = usecase_summary
        self._summary_phrase = summary_phrase
        self._summary_text = summary_text

    @property
    def headline(self) -> str:
        return self._headline

    @property
    def author(self) -> str:
        return self._author

    @property
    def date(self) -> str:
        return self._date

    @property
    def system_title(self) -> str:
        return self._system_title

    @property
    def system_summary(self) -> str:
        return self._system_summary

    @property
    def usecase_title(self) -> str:
        return self._usecase_title

    @property
    def usecase_summary(self) -> str:
        return self._usecase_summary

    @property
    def summary_phrase(self) -> str:
        return self._summary_phrase

    @property
    def summary_text(self) -> str:
        return self._summary_text

    def _repr_html_(self) -> str:
        template_file = importlib.resources.files(templates) / "TitleCard.template.html"
        with template_file.open("r") as fin:
            template = string.Template(fin.read())

        return template.substitute(
            headline=self._headline,
            author=self._author,
            date=self._date,
            system_title=self._system_title,
            system_summary=self._system_summary,
            usecase_title=self._usecase_title,
            usecase_summary=self._usecase_summary,
            summary_phrase=self._summary_phrase,
            summary_text=self._summary_text,
        )



# SPDX-FileCopyrightText: 2024 UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

# mypy: disable-error-code="import-untyped"
import abc
import random
import re
from typing import Any, Iterable, Optional

import pyarrow
import pydantic

from dyff.schema import ids
from dyff.schema.adapters import Adapter
from dyff.schema.dataset import arrow
from dyff.schema.dataset.text import Text
from dyff.schema.platform import DataSchema, DyffSchemaBaseModel, InferenceInterface


class InferenceServiceMock(abc.ABC):
    @property
    @abc.abstractmethod
    def schema(self) -> DataSchema:
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def interface(self) -> InferenceInterface:
        raise NotImplementedError()

    @abc.abstractmethod
    def infer(self, endpoint: str, request: dict[str, Any]) -> list[dict[str, Any]]:
        raise NotImplementedError()


class InferenceSessionClientMock:
    def __init__(
        self,
        service: InferenceServiceMock,
        *,
        inference_endpoint: str,
        input_adapter: Optional[Adapter] = None,
        output_adapter: Optional[Adapter] = None,
    ):
        self._service = service
        self._inference_endpoint = inference_endpoint
        self._input_adapter = input_adapter
        self._output_adapter = output_adapter

    def infer(self, body: dict[str, Any]) -> list[dict[str, Any]]:
        def once(x):
            if isinstance(x, list):
                yield from x
            else:
                yield x

        adapted_input = once(body)
        if self._input_adapter is not None:
            adapted_input = self._input_adapter(adapted_input)

        request_body = None
        for i, x in enumerate(adapted_input):
            if i > 0:
                raise ValueError("adapted input should contain exactly one element")
            request_body = x
        if request_body is None:
            raise ValueError("adapted input should contain exactly one element")

        json_response = self._service.infer(
            self._service.interface.endpoint, request_body
        )
        adapted_response = once(json_response)
        if self._output_adapter is not None:
            adapted_response = self._output_adapter(adapted_response)
        return list(adapted_response)

    def run_evaluation(
        self,
        input_dataset: pyarrow.dataset.Dataset,
        *,
        id: str,
        replications: int = 1,
    ) -> Iterable[pyarrow.RecordBatch]:
        interface = self._service.interface
        replication_ids = [ids.replication_id(id, i) for i in range(replications)]
        feature_schema = arrow.decode_schema(interface.outputSchema.arrowSchema)

        def output_generator() -> Iterable[pyarrow.RecordBatch]:
            for input_batch in input_dataset.to_batches():
                output_batch: list[dict[str, Any]] = []
                for item in input_batch.to_pylist():
                    index = item["_index_"]
                    for replication in replication_ids:
                        responses = self.infer(item)
                        for i, response in enumerate(responses):
                            response["_response_index_"] = i
                        response_record = {
                            "_replication_": replication,
                            "_index_": index,
                            "responses": responses,
                        }
                        output_batch.append(response_record)
                yield pyarrow.RecordBatch.from_pylist(  # type: ignore[attr-defined]
                    output_batch, schema=feature_schema
                )

        yield from output_generator()


class TextCompletionOutputSchema(DyffSchemaBaseModel):
    text: list[str] = pydantic.Field(
        default_factory=list, description="Text completions"
    )


class TextCompletion(InferenceServiceMock):
    def __init__(self):
        self._schema = DataSchema.make_output_schema(Text)
        self._num_responses = 2

    @property
    def schema(self) -> DataSchema:
        return self._schema

    @property
    def interface(self) -> InferenceInterface:
        return InferenceInterface(endpoint="generate", outputSchema=self.schema)

    def infer(self, endpoint: str, request: dict[str, Any]) -> list[dict[str, Any]]:
        if endpoint != "generate":
            raise ValueError(f"unknown endpoint '{endpoint}'")
        return self.generate(request)

    def generate(self, x: dict[str, Any]) -> list[dict[str, Any]]:
        text = x["text"]

        responses = []
        for i in range(self._num_responses):
            c = random.randrange(5)
            if c == 0:
                response = " I'm sorry, Dave. I'm afraid I can't do that."
            elif c == 1:
                response = " All work and no play makes Jack a dull boy."
            elif c == 2:
                prompt_words: list[str] = re.split(r"\s", text.strip())
                if len(prompt_words) > 0 and random.randrange(2):
                    n = min(len(prompt_words), 3)
                    echo = " ".join(w.upper() for w in prompt_words[-n:])
                    response = f" You should write a book, Fry! People need to know about the {echo}."
                else:
                    response = "POETIC IMAGE #35 NOT FOUND."

            elif c == 3:
                response = " it was the worst of times."
            else:
                response = " it was the blurst of times."
            responses.append(response)
        return [{"text": text + response} for response in responses]
